use miette::Diagnostic;
use serde::{Deserialize, Serialize};
use serde_repr::Deserialize_repr;
use std::collections::HashMap;
use thiserror::Error;

pub(crate) mod internal;

#[derive(Debug, Clone, Deserialize, Error, Diagnostic)]
#[error("Failed to register bridge: {description}")]
pub struct ApiError {
    #[serde(rename = "type")]
    pub err_type: ApiErrorType,
    pub address: String,
    pub description: String,
}

#[derive(Debug, Clone, Deserialize_repr)]
#[repr(u8)]
pub enum ApiErrorType {
    InvalidValue = 7,
    LinkButtonNotPressed = 101,
    NotAvailable = 2,
    ParameterNotAvailable = 6,
    ParameterNotModifiable = 201,
    ResourceNotAvailable = 3,
    UnauthorizedUser = 1,
}

#[derive(Debug, Clone, Deserialize)]
pub struct Light {
    pub state: State,
    pub swupdate: SwUpdate,
    #[serde(rename = "type")]
    pub light_type: String,
    pub name: String,
    pub modelid: String,
    pub manufacturername: String,
    pub productname: String,
    pub capabilities: LightCapabilities,
    pub config: LightConfig,
    pub uniqueid: String,
    pub swversion: String,
    pub swconfigid: Option<String>,
    pub productid: Option<String>,
}

#[derive(Debug, Clone, Deserialize)]
pub struct LightConfig {
    pub archetype: LightArchetype,
    pub function: String,
    pub direction: String,
    pub startup: LightStartupConfig,
}

#[derive(Debug, Clone, Deserialize)]
#[serde(rename_all = "lowercase")]
pub enum LightArchetype {
    Bollard,
    Candlebulb,
    Ceilinground,
    Ceilingsquare,
    Christmastree,
    Classicbulb,
    Doublespot,
    Flexiblelamp,
    Floodbulb,
    Floorlantern,
    Floorshade,
    Groundspot,
    Huebloom,
    Huecentris,
    Huego,
    Hueiris,
    Huelightstrip,
    Huelightstriptv,
    Hueplay,
    Pendantlong,
    Pendantround,
    Recessedceiling,
    Recessedfloor,
    Singlespot,
    Spotbulb,
    Sultanbulb,
    Tableshade,
    Tablewash,
    Vintagebulb,
    Walllantern,
    Wallshade,
    Wallspot,

    #[serde(other)]
    Other,
}

#[derive(Debug, Clone, Deserialize)]
pub struct LightStartupConfig {
    // TODO: find variants and use enum
    pub mode: String,
    pub configured: bool,
}

#[derive(Debug, Clone, Deserialize)]
pub struct LightCapabilities {
    #[serde(default)]
    pub certified: bool,
    pub control: LightControlCapabilities,
    pub streaming: Option<LightStreamingCapabilities>,
}

#[derive(Debug, Clone, Deserialize)]
pub struct LightStreamingCapabilities {
    pub renderer: bool,
    pub proxy: bool,
}

#[derive(Debug, Clone, Deserialize)]
pub struct LightControlCapabilities {
    pub mindimlevel: Option<u32>,
    pub maxlumen: Option<u32>,
    pub colorgamuttype: Option<String>,
    pub colorgamut: Option<[[f32; 2]; 3]>,
    pub ct: Option<LightCtCapabilities>,
}

#[derive(Debug, Clone, Deserialize)]
pub struct LightCtCapabilities {
    pub min: u16,
    pub max: u16,
}

#[derive(Debug, Clone, Deserialize)]
pub struct SwUpdate {
    pub state: String,
    pub lastinstall: String,
}

#[derive(Debug, Clone, Deserialize)]
pub struct State {
    pub on: bool,
    pub bri: u8,
    pub hue: u16,
    pub sat: u8,
    pub effect: LightEffect,
    pub xy: [f32; 2],
    pub ct: Option<u16>,
    pub alert: LightAlert,
    pub colormode: ColorMode,
    // TODO: find out what variants this has and replace with enum
    pub mode: Option<String>,
    pub reachable: Option<bool>,
}

#[derive(Debug, Clone, Deserialize)]
#[serde(rename_all = "lowercase")]
pub enum LightAlert {
    Select,
    LSelect,
    None,
}

#[derive(Debug, Default, Clone, Serialize)]
pub struct StateUpdate {
    #[serde(skip_serializing_if = "Option::is_none")]
    pub on: Option<bool>,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub bri: Option<u8>,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub hue: Option<u16>,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub sat: Option<u8>,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub effect: Option<LightEffect>,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub xy: Option<[f32; 2]>,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub colormode: Option<ColorMode>,
    // TODO: find out what variants this has and replace with enum
    #[serde(skip_serializing_if = "Option::is_none")]
    pub mode: Option<String>,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub reachable: Option<bool>,
}

#[derive(Debug, Clone, Deserialize, Serialize)]
#[serde(rename_all = "lowercase")]
pub enum LightEffect {
    None,
    ColorLoop,
}

#[derive(Debug, Clone, Deserialize, Serialize)]
pub enum ColorMode {
    #[serde(rename = "hs")]
    HueSatBri,
    #[serde(rename = "xy")]
    Xy,
    #[serde(rename = "ct")]
    ColorTemp,
}

#[derive(Debug, Clone, Deserialize)]
pub struct UpdateStatus {
    pub success: Option<HashMap<String, serde_json::Value>>,
    pub error: Option<ApiError>,
}

#[derive(Debug, Clone, Deserialize)]
pub struct Group {
    pub name: String,
    pub lights: Vec<String>,
    #[serde(rename = "type")]
    pub group_type: GroupType,
    pub state: Grouptate,
    #[serde(default)]
    pub recycle: bool,
    pub class: GroupClass,
    pub action: State,
    pub stream: Option<GroupStream>,
    pub locations: Option<HashMap<String, [f32; 3]>>,
}

#[derive(Debug, Clone, Deserialize)]
pub enum GroupType {
    Room,
    Entertainment,
}

#[derive(Debug, Clone, Deserialize)]
pub struct Grouptate {
    pub all_on: bool,
    pub any_on: bool,
}

#[derive(Debug, Clone, Deserialize)]
pub enum GroupClass {
    Other,
}

#[derive(Debug, Clone, Deserialize)]
pub struct GroupStream {
    // TODO: enumify
    pub proxymode: String,
    pub proxynode: String,
    pub active: bool,
    pub owner: Option<String>,
}
