use std::{collections::HashMap, sync::Arc};

use crate::api::{
    internal::{RegisterReq, RegisterResponse, RegisterSuccess},
    ApiError,
    Group,
    Light,
    UpdateStatus,
};
use miette::Diagnostic;
use reqwest::Client;
use thiserror::Error;
use url::Url;

pub use crate::api::StateUpdate;

pub type BridgeResult<T> = Result<T, BridgeError>;

#[derive(Error, Debug, Diagnostic)]
pub enum BridgeError {
    #[error("Error sending HTTP request")]
    ReqwestError(#[from] reqwest::Error),
    #[error("Error while serializing JSON data")]
    JsonError(#[from] serde_json::Error),
    #[error("Error parsing URL")]
    UrlParseError(#[from] url::ParseError),
    #[error("Response from bridge had invalid data.")]
    ResponseFormatError,
}

#[derive(Debug)]
pub struct Bridge {
    pub endpoint: Url,
    pub username: String,
    http: Arc<Client>,
}

#[derive(Debug)]
pub enum ActionStatus<T> {
    Success(T),
    SuccessAndError(T, ApiError),
    Error(ApiError),
}

impl<T> ActionStatus<T> {
    pub fn to_result(self) -> Result<T, ApiError> {
        match self {
            Self::Success(b) => Ok(b),
            Self::SuccessAndError(b, _) => Ok(b),
            Self::Error(e) => Err(e),
        }
    }
}

impl Bridge {
    /// tries to register a new user at the given bridge endpoint with a given
    /// username.
    pub async fn register(
        http: Arc<Client>,
        devicetype: &str,
        addr: Url,
    ) -> BridgeResult<ActionStatus<Bridge>> {
        let req = serde_json::to_vec(&RegisterReq { devicetype })?;

        let response = http
            .post(addr.join("api")?)
            .body(req)
            .send()
            .await?
            .bytes()
            .await?;

        // no clue why the response is an array, and not just a single object
        let mut response = serde_json::from_slice::<Vec<RegisterResponse>>(&response)?;
        let response = response.pop().ok_or(BridgeError::ResponseFormatError)?;

        match response {
            RegisterResponse {
                error: Some(error),
                success: Some(RegisterSuccess { username }),
            } => Ok(ActionStatus::SuccessAndError(
                Bridge::with_http(username, addr, http),
                error,
            )),
            RegisterResponse {
                error: None,
                success: Some(RegisterSuccess { username }),
            } => Ok(ActionStatus::Success(Bridge::with_http(
                username, addr, http,
            ))),
            RegisterResponse {
                error: Some(error),
                success: None,
            } => Ok(ActionStatus::Error(error)),
            RegisterResponse {
                error: None,
                success: None,
            } => Err(BridgeError::ResponseFormatError),
        }
    }

    pub fn new(username: impl ToString, endpoint: Url) -> Self {
        Self::with_http(username, endpoint, Arc::new(Client::new()))
    }

    pub fn with_http(username: impl ToString, endpoint: Url, http: Arc<Client>) -> Self {
        Self {
            endpoint,
            username: username.to_string(),
            http,
        }
    }

    fn api_endpoint(&self) -> Result<Url, url::ParseError> {
        self.endpoint
            .join("api/")?
            .join(&(self.username.clone() + "/"))
    }

    pub async fn get_lights(&self) -> BridgeResult<HashMap<String, Light>> {
        let resp = self
            .http
            .get(self.api_endpoint()?.join("lights/")?)
            .send()
            .await?
            .bytes()
            .await?;

        Ok(serde_json::from_slice(&resp)?)
    }

    async fn update_endpoint(
        &self,
        endpoint: &str,
        update: StateUpdate,
    ) -> BridgeResult<Vec<ActionStatus<HashMap<String, serde_json::Value>>>> {
        let payload = serde_json::to_vec(&update)?;
        let res = self
            .http
            .put(self.api_endpoint()?.join(endpoint)?)
            .body(payload)
            .send()
            .await?
            .bytes()
            .await?;

        let res = serde_json::from_slice::<Vec<UpdateStatus>>(&res)?;

        let mut results = vec![];

        for status in res {
            results.push(match status {
                UpdateStatus {
                    success: Some(s),
                    error: None,
                } => ActionStatus::Success(s),
                UpdateStatus {
                    success: None,
                    error: Some(e),
                } => ActionStatus::Error(e),
                UpdateStatus {
                    success: Some(s),
                    error: Some(e),
                } => ActionStatus::SuccessAndError(s, e),
                UpdateStatus {
                    success: None,
                    error: None,
                } => return Err(BridgeError::ResponseFormatError),
            });
        }

        Ok(results)
    }

    pub async fn update_light(
        &self,
        id: &str,
        update: StateUpdate,
    ) -> BridgeResult<Vec<ActionStatus<HashMap<String, serde_json::Value>>>> {
        self.update_endpoint(&format!("lights/{}/state", id), update)
            .await
    }

    pub async fn get_groups(&self) -> BridgeResult<HashMap<String, Group>> {
        let res = self
            .http
            .get(self.api_endpoint()?.join("groups")?)
            .send()
            .await?
            .bytes()
            .await?;

        Ok(serde_json::from_slice(&res)?)
    }

    pub async fn update_group(
        &self,
        id: &str,
        update: StateUpdate,
    ) -> BridgeResult<Vec<ActionStatus<HashMap<String, serde_json::Value>>>> {
        self.update_endpoint(&format!("groups/{}/action", id), update)
            .await
    }
}
