use miette::Diagnostic;
use std::{io, sync::Arc, time::Duration};
use thiserror::Error;
use tokio_stream::StreamExt;
use url::Url;

use ssdp_client::SearchTarget;

pub mod api;
pub mod bridge;

/// an error that can occur while discovering a hue bridge
#[derive(Error, Debug, Diagnostic)]
pub enum DiscoverError {
    #[error(transparent)]
    IoError(#[from] io::Error),
    #[error(transparent)]
    SsdpError(#[from] ssdp_client::Error),
    #[error(transparent)]
    ReqwestError(#[from] reqwest::Error),
    #[error(transparent)]
    UrlError(#[from] url::ParseError),
}

/// Looks for hue bridges in the local network using the SSDP protocol.
/// A reqwest client is required, to fetch the device's description, which is
/// needed to make sure a device is a hue bridge. If a bridge is found, it's URL
/// is returned.
pub async fn discover_bridge(
    reqwest_client: Arc<reqwest::Client>,
    timeout: Duration,
) -> Result<Option<Url>, DiscoverError> {
    let mut ssdp = ssdp_client::search(&SearchTarget::All, timeout, 2).await?;

    while let Some(resp) = ssdp.next().await {
        let resp = resp?;

        if resp.location().ends_with("/description.xml") {
            let mut url = Url::parse(resp.location())?;

            // this check is needed so we don't think that all devices that use
            // description.xml are hue bridges.
            if reqwest_client
                .get(url.clone())
                .send()
                .await?
                .text()
                .await?
                .contains("hue")
            {
                url.set_path("");
                return Ok(Some(url));
            }
        }
    }

    Ok(None)
}

#[derive(Error, Debug, Diagnostic)]
pub enum Error {
    #[error("Reqwest error")]
    ReqwestError(#[from] reqwest::Error),
    #[error("JSON error")]
    JsonError(#[from] serde_json::Error),
}
